@@ -24,6 +24,7 @@
 
 #include "soup-websocket-connection.h"
 #include "soup-enum-types.h"
+#include "soup-io-stream.h"
 #include "soup-uri.h"
 
 /*
@@ -147,6 +148,8 @@ struct _SoupWebsocketConnectionPrivate {
 };
 
 #define MAX_INCOMING_PAYLOAD_SIZE_DEFAULT   128 * 1024
+#define READ_BUFFER_SIZE 1024
+#define MASK_LENGTH 4
 
 G_DEFINE_TYPE_WITH_PRIVATE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
 
@@ -155,6 +158,87 @@ static void queue_frame (SoupWebsocketConnection *self, SoupWebsocketQueueFlags
 
 static void protocol_error_and_close (SoupWebsocketConnection *self);
 
+static gboolean on_web_socket_input (GObject *pollable_stream,
+				     gpointer user_data);
+static gboolean on_web_socket_output (GObject *pollable_stream,
+				      gpointer user_data);
+
+/* Code below is based on g_utf8_validate() implementation,
+ * but handling NULL characters as valid, as expected by
+ * WebSockets and compliant with RFC 3629.
+ */
+#define VALIDATE_BYTE(mask, expect)                             \
+        G_STMT_START {                                          \
+          if (G_UNLIKELY((*(guchar *)p & (mask)) != (expect)))  \
+                  return FALSE;                                 \
+        } G_STMT_END
+
+/* see IETF RFC 3629 Section 4 */
+static gboolean
+utf8_validate (const char *str,
+               gsize max_len)
+
+{
+        const gchar *p;
+
+        for (p = str; ((p - str) < max_len); p++) {
+                if (*(guchar *)p < 128)
+                        /* done */;
+                else {
+                        if (*(guchar *)p < 0xe0) { /* 110xxxxx */
+                                if (G_UNLIKELY (max_len - (p - str) < 2))
+                                        return FALSE;
+
+                                if (G_UNLIKELY (*(guchar *)p < 0xc2))
+                                        return FALSE;
+                        } else {
+                                if (*(guchar *)p < 0xf0) { /* 1110xxxx */
+                                        if (G_UNLIKELY (max_len - (p - str) < 3))
+                                                return FALSE;
+
+                                        switch (*(guchar *)p++ & 0x0f) {
+                                        case 0:
+                                                VALIDATE_BYTE(0xe0, 0xa0); /* 0xa0 ... 0xbf */
+                                                break;
+                                        case 0x0d:
+                                                VALIDATE_BYTE(0xe0, 0x80); /* 0x80 ... 0x9f */
+                                                break;
+                                        default:
+                                                VALIDATE_BYTE(0xc0, 0x80); /* 10xxxxxx */
+                                        }
+                                } else if (*(guchar *)p < 0xf5) { /* 11110xxx excluding out-of-range */
+                                        if (G_UNLIKELY (max_len - (p - str) < 4))
+                                                return FALSE;
+
+                                        switch (*(guchar *)p++ & 0x07) {
+                                        case 0:
+                                                VALIDATE_BYTE(0xc0, 0x80); /* 10xxxxxx */
+                                                if (G_UNLIKELY((*(guchar *)p & 0x30) == 0))
+                                                        return FALSE;
+                                                break;
+                                        case 4:
+                                                VALIDATE_BYTE(0xf0, 0x80); /* 0x80 ... 0x8f */
+                                                break;
+                                        default:
+                                                VALIDATE_BYTE(0xc0, 0x80); /* 10xxxxxx */
+                                        }
+                                        p++;
+                                        VALIDATE_BYTE(0xc0, 0x80); /* 10xxxxxx */
+                                } else {
+                                        return FALSE;
+                                }
+                        }
+
+                        p++;
+                        VALIDATE_BYTE(0xc0, 0x80); /* 10xxxxxx */
+                }
+        }
+
+        return TRUE;
+}
+
+#undef VALIDATE_BYTE
+
 static void
 frame_free (gpointer data)
 {
@@ -207,7 +291,20 @@ on_iostream_closed (GObject *source,
 }
 
 static void
-stop_input (SoupWebsocketConnection *self)
+soup_websocket_connection_start_input_source (SoupWebsocketConnection *self)
+{
+	SoupWebsocketConnectionPrivate *pv = self->pv;
+
+	if (pv->input_source)
+		return;
+
+	pv->input_source = g_pollable_input_stream_create_source (pv->input, NULL);
+	g_source_set_callback (pv->input_source, (GSourceFunc)on_web_socket_input, self, NULL);
+	g_source_attach (pv->input_source, pv->main_context);
+}
+
+static void
+soup_websocket_connection_stop_input_source (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
@@ -220,7 +317,20 @@ stop_input (SoupWebsocketConnection *self)
 }
 
 static void
-stop_output (SoupWebsocketConnection *self)
+soup_websocket_connection_start_output_source (SoupWebsocketConnection *self)
+{
+	SoupWebsocketConnectionPrivate *pv = self->pv;
+
+	if (pv->output_source)
+		return;
+
+	pv->output_source = g_pollable_output_stream_create_source (pv->output, NULL);
+	g_source_set_callback (pv->output_source, (GSourceFunc)on_web_socket_output, self, NULL);
+	g_source_attach (pv->output_source, pv->main_context);
+}
+
+static void
+soup_websocket_connection_stop_output_source (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
@@ -265,8 +375,8 @@ close_io_stream (SoupWebsocketConnection *self)
 	close_io_stop_timeout (self);
 
 	if (!pv->io_closing) {
-		stop_input (self);
-		stop_output (self);
+		soup_websocket_connection_stop_input_source (self);
+		soup_websocket_connection_stop_output_source (self);
 		pv->io_closing = TRUE;
 		g_debug ("closing io stream");
 		g_io_stream_close_async (pv->io_stream, G_PRIORITY_DEFAULT,
@@ -281,12 +391,17 @@ shutdown_wr_io_stream (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	GSocket *socket;
+	GIOStream *base_iostream;
 	GError *error = NULL;
 
-	stop_output (self);
+	soup_websocket_connection_stop_output_source (self);
 
-	if (G_IS_SOCKET_CONNECTION (pv->io_stream)) {
-		socket = g_socket_connection_get_socket (G_SOCKET_CONNECTION (pv->io_stream));
+	base_iostream = SOUP_IS_IO_STREAM (pv->io_stream) ?
+		soup_io_stream_get_base_iostream (SOUP_IO_STREAM (pv->io_stream)) :
+		pv->io_stream;
+
+	if (G_IS_SOCKET_CONNECTION (base_iostream)) {
+		socket = g_socket_connection_get_socket (G_SOCKET_CONNECTION (base_iostream));
 		g_socket_shutdown (socket, FALSE, TRUE, &error);
 		if (error != NULL) {
 			g_debug ("error shutting down io stream: %s", error->message);
@@ -349,8 +464,7 @@ send_message (SoupWebsocketConnection *self,
 	GByteArray *bytes;
 	gsize frame_len;
 	guint8 *outer;
-	guint8 *mask = 0;
-	guint8 *at;
+	guint8 mask_offset;
 
 	if (!(soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN)) {
 		g_debug ("Ignoring message since the connection is closed or is closing");
@@ -366,6 +480,7 @@ send_message (SoupWebsocketConnection *self,
 		if (length > 125) {
 			g_warning ("WebSocket control message payload exceeds size limit");
 			protocol_error_and_close (self);
+			g_byte_array_free (bytes, TRUE);
 			return;
 		}
 
@@ -401,17 +516,17 @@ send_message (SoupWebsocketConnection *self,
 	 * probably a client somewhere that's not expecting it.
 	 */
 	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT) {
+		guint32 rnd = g_random_int ();
 		outer[1] |= 0x80;
-		mask = outer + bytes->len;
-		* ((guint32 *)mask) = g_random_int ();
-		bytes->len += 4;
+		mask_offset = bytes->len;
+		memcpy (outer + mask_offset, &rnd, sizeof (rnd));
+		bytes->len += MASK_LENGTH;
 	}
 
-	at = bytes->data + bytes->len;
 	g_byte_array_append (bytes, data, length);
 
 	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT)
-		xor_with_mask (mask, at, length);
+		xor_with_mask (bytes->data + mask_offset, bytes->data + mask_offset + MASK_LENGTH, length);
 
 	frame_len = bytes->len;
 	queue_frame (self, flags, g_byte_array_free (bytes, FALSE),
@@ -535,9 +650,6 @@ too_big_error_and_close (SoupWebsocketConnection *self,
 		 self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER ? "server" : "client",
 	         payload_len, self->pv->max_incoming_payload_size);
 	emit_error_and_close (self, error, TRUE);
-
-	/* The input is in an invalid state now */
-	stop_input (self);
 }
 
 static void
@@ -577,6 +689,10 @@ close_connection (SoupWebsocketConnection *self,
 			         code);
 		}
 		break;
+	case SOUP_WEBSOCKET_CLOSE_NO_STATUS:
+		/* This is special case to send a close message with no body */
+		code = 0;
+		break;
 	default:
 		if (code < 3000) {
 			g_debug ("Wrong closing code %d received", code);
@@ -591,7 +707,7 @@ close_connection (SoupWebsocketConnection *self,
 		g_debug ("responding to close request");
 
 	flags = 0;
-	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && pv->close_received)
+	if (pv->close_received)
 		flags |= SOUP_WEBSOCKET_QUEUE_LAST;
 	send_close (self, flags, code, data);
 	close_io_after_timeout (self);
@@ -612,6 +728,7 @@ receive_close (SoupWebsocketConnection *self,
 	switch (len) {
 	case 0:
 		/* Send a clean close when having an empty payload */
+		pv->peer_close_code = SOUP_WEBSOCKET_CLOSE_NO_STATUS;
 		close_connection (self, 1000, NULL);
 		return;
 	case 1:
@@ -628,7 +745,7 @@ receive_close (SoupWebsocketConnection *self,
 		data += 2;
 		len -= 2;
 		
-		if (!g_utf8_validate ((char *)data, len, NULL)) {
+		if (!utf8_validate ((const char *)data, len)) {
 			g_debug ("received non-UTF8 close data: %d '%.*s' %d", (int)len, (int)len, (char *)data, (int)data[0]);
 			protocol_error_and_close (self);
 			return;
@@ -689,6 +806,9 @@ process_contents (SoupWebsocketConnection *self,
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	GBytes *message;
 
+	if (pv->close_sent && pv->close_received)
+		return;
+
 	if (control) {
 		/* Control frames must never be fragmented */
 		if (!fin) {
@@ -773,9 +893,8 @@ process_contents (SoupWebsocketConnection *self,
 		/* Actually deliver the message? */
 		if (fin) {
 			if (pv->message_opcode == 0x01 &&
-			    !g_utf8_validate((char *)pv->message_data->data,
-			                     pv->message_data->len,
-			                     NULL)) {
+			    !utf8_validate((const char *)pv->message_data->data,
+					   pv->message_data->len)) {
 
 				g_debug ("received invalid non-UTF8 text data");
 
@@ -830,20 +949,52 @@ process_frame (SoupWebsocketConnection *self)
 	opcode = header[0] & 0x0f;
 	masked = ((header[1] & 0x80) != 0);
 
+	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT && masked) {
+		/* A server MUST NOT mask any frames that it sends to the client.
+		 * A client MUST close a connection if it detects a masked frame.
+		 */
+		g_debug ("A server must not mask any frames that it sends to the client.");
+		protocol_error_and_close (self);
+		return FALSE;
+	}
+
+	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && !masked) {
+		/* The server MUST close the connection upon receiving a frame
+		 * that is not masked.
+		 */
+		g_debug ("The client should always mask frames");
+		protocol_error_and_close (self);
+                return FALSE;
+        }
+
 	/* We do not support extensions, reserved bits must be 0 */
 	if (header[0] & 0x70) {
 		protocol_error_and_close (self);
+		return FALSE;
 	}
 
 	switch (header[1] & 0x7f) {
 	case 126:
+		/* If 126, the following 2 bytes interpreted as a 16-bit
+		 * unsigned integer are the payload length.
+		 */
 		at = 4;
 		if (len < at)
 			return FALSE; /* need more data */
 		payload_len = (((guint16)header[2] << 8) |
 			       ((guint16)header[3] << 0));
+
+		/* The minimal number of bytes MUST be used to encode the length. */
+		if (payload_len <= 125) {
+			protocol_error_and_close (self);
+			return FALSE;
+		}
 		break;
 	case 127:
+		/* If 127, the following 8 bytes interpreted as a 64-bit
+		 * unsigned integer (the most significant bit MUST be 0)
+		 * are the payload length.
+		 */
 		at = 10;
 		if (len < at)
 			return FALSE; /* need more data */
@@ -855,6 +1006,12 @@ process_frame (SoupWebsocketConnection *self)
 			       ((guint64)header[7] << 16) |
 			       ((guint64)header[8] << 8) |
 			       ((guint64)header[9] << 0));
+
+		/* The minimal number of bytes MUST be used to encode the length. */
+		if (payload_len <= G_MAXUINT16) {
+			protocol_error_and_close (self);
+			return FALSE;
+		}
 		break;
 	default:
 		payload_len = header[1] & 0x7f;
@@ -902,32 +1059,31 @@ process_incoming (SoupWebsocketConnection *self)
 		;
 }
 
-static gboolean
-on_web_socket_input (GObject *pollable_stream,
-		     gpointer user_data)
+static void
+soup_websocket_connection_read (SoupWebsocketConnection *self)
 {
-	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	GError *error = NULL;
 	gboolean end = FALSE;
 	gssize count;
 	gsize len;
 
+	soup_websocket_connection_stop_input_source (self);
+
 	do {
 		len = pv->incoming->len;
-		g_byte_array_set_size (pv->incoming, len + 1024);
+		g_byte_array_set_size (pv->incoming, len + READ_BUFFER_SIZE);
 
 		count = g_pollable_input_stream_read_nonblocking (pv->input,
 								  pv->incoming->data + len,
-								  1024, NULL, &error);
-
+								  READ_BUFFER_SIZE, NULL, &error);
 		if (count < 0) {
 			if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
 				g_error_free (error);
 				count = 0;
 			} else {
 				emit_error_and_close (self, error, TRUE);
-				return TRUE;
+				return;
 			}
 		} else if (count == 0) {
 			end = TRUE;
@@ -947,16 +1103,25 @@ on_web_socket_input (GObject *pollable_stream,
 		}
 
 		close_io_stream (self);
+		return;
 	}
 
-	return TRUE;
+	if (!pv->io_closing)
+		soup_websocket_connection_start_input_source (self);
 }
 
 static gboolean
-on_web_socket_output (GObject *pollable_stream,
-		      gpointer user_data)
+on_web_socket_input (GObject *pollable_stream,
+		     gpointer user_data)
+{
+	soup_websocket_connection_read (SOUP_WEBSOCKET_CONNECTION (user_data));
+
+	return G_SOURCE_REMOVE;
+}
+
+static void
+soup_websocket_connection_write (SoupWebsocketConnection *self)
 {
-	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	const guint8 *data;
 	GError *error = NULL;
@@ -964,19 +1129,18 @@ on_web_socket_output (GObject *pollable_stream,
 	gssize count;
 	gsize len;
 
+	soup_websocket_connection_stop_output_source (self);
+
 	if (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_CLOSED) {
 		g_debug ("Ignoring message since the connection is closed");
-		stop_output (self);
-		return TRUE;
+		return;
 	}
 
 	frame = g_queue_peek_head (&pv->outgoing);
 
 	/* No more frames to send */
-	if (frame == NULL) {
-		stop_output (self);
-		return TRUE;
-	}
+	if (frame == NULL)
+		return;
 
 	data = g_bytes_get_data (frame->data, &len);
 	g_assert (len > 0);
@@ -996,7 +1160,7 @@ on_web_socket_output (GObject *pollable_stream,
 			frame->pending = TRUE;
 		} else {
 			emit_error_and_close (self, error, TRUE);
-			return FALSE;
+			return;
 		}
 	}
 
@@ -1014,23 +1178,21 @@ on_web_socket_output (GObject *pollable_stream,
 			}
 		}
 		frame_free (frame);
+
+		if (g_queue_is_empty (&pv->outgoing))
+			return;
 	}
 
-	return TRUE;
+	soup_websocket_connection_start_output_source (self);
 }
 
-static void
-start_output (SoupWebsocketConnection *self)
+static gboolean
+on_web_socket_output (GObject *pollable_stream,
+		      gpointer user_data)
 {
-	SoupWebsocketConnectionPrivate *pv = self->pv;
+	soup_websocket_connection_write (SOUP_WEBSOCKET_CONNECTION (user_data));
 
-	if (pv->output_source)
-		return;
-
-	g_debug ("starting output source");
-	pv->output_source = g_pollable_output_stream_create_source (pv->output, NULL);
-	g_source_set_callback (pv->output_source, (GSourceFunc)on_web_socket_output, self, NULL);
-	g_source_attach (pv->output_source, pv->main_context);
+	return G_SOURCE_REMOVE;
 }
 
 static void
@@ -1071,7 +1233,7 @@ queue_frame (SoupWebsocketConnection *self,
 		g_queue_push_tail (&pv->outgoing, frame);
 	}
 
-	start_output (self);
+	soup_websocket_connection_write (self);
 }
 
 static void
@@ -1096,9 +1258,7 @@ soup_websocket_connection_constructed (GObject *object)
 	pv->output = G_POLLABLE_OUTPUT_STREAM (os);
 	g_return_if_fail (g_pollable_output_stream_can_poll (pv->output));
 
-	pv->input_source = g_pollable_input_stream_create_source (pv->input, NULL);
-	g_source_set_callback (pv->input_source, (GSourceFunc)on_web_socket_input, self, NULL);
-	g_source_attach (pv->input_source, pv->main_context);
+	soup_websocket_connection_start_input_source (self);
 }
 
 static void
@@ -1695,7 +1855,9 @@ soup_websocket_connection_get_close_data (SoupWebsocketConnection *self)
  * @self: the WebSocket
  * @text: the message contents
  *
- * Send a text (UTF-8) message to the peer.
+ * Send a %NULL-terminated text (UTF-8) message to the peer. If you need
+ * to send text messages containing %NULL characters use
+ * soup_websocket_connection_send_message() instead.
  *
  * The message is queued to be sent and will be sent when the main loop
  * is run.
@@ -1713,7 +1875,7 @@ soup_websocket_connection_send_text (SoupWebsocketConnection *self,
 	g_return_if_fail (text != NULL);
 
 	length = strlen (text);
-	g_return_if_fail (g_utf8_validate (text, length, NULL));
+        g_return_if_fail (utf8_validate (text, length));
 
 	send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, 0x01, (const guint8 *) text, length);
 }
@@ -1721,10 +1883,10 @@ soup_websocket_connection_send_text (SoupWebsocketConnection *self,
 /**
  * soup_websocket_connection_send_binary:
  * @self: the WebSocket
- * @data: (array length=length) (element-type guint8): the message contents
+ * @data: (array length=length) (element-type guint8) (nullable): the message contents
  * @length: the length of @data
  *
- * Send a binary message to the peer.
+ * Send a binary message to the peer. If @length is 0, @data may be %NULL.
  *
  * The message is queued to be sent and will be sent when the main loop
  * is run.
@@ -1738,7 +1900,7 @@ soup_websocket_connection_send_binary (SoupWebsocketConnection *self,
 {
 	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
 	g_return_if_fail (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN);
-	g_return_if_fail (data != NULL);
+	g_return_if_fail (data != NULL || length == 0);
 
 	send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, 0x02, data, length);
 }
@@ -1756,6 +1918,8 @@ soup_websocket_connection_send_binary (SoupWebsocketConnection *self,
  * main loop runs.
  *
  * The @code and @data are sent to the peer along with the close request.
+ * If @code is %SOUP_WEBSOCKET_CLOSE_NO_STATUS a close message with no body
+ * (without code and data) is sent.
  * Note that the @data must be UTF-8 valid.
  *
  * Since: 2.50
@@ -1771,8 +1935,7 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 	pv = self->pv;
 	g_return_if_fail (!pv->close_sent);
 
-	g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_NO_STATUS &&
-			  code != SOUP_WEBSOCKET_CLOSE_ABNORMAL &&
+	g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_ABNORMAL &&
 			  code != SOUP_WEBSOCKET_CLOSE_TLS_HANDSHAKE);
 	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER)
 		g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_NO_EXTENSION);
